=head1 NAME

Tails::IUK::Utils - utilities for Tails IUK

=cut

package Tails::IUK::Utils;

use strict;
use warnings FATAL => 'all';
use 5.10.1;



use Exporter;
our @ISA = qw{Exporter};
our @EXPORT = (
    qw{extract_file_from_iso extract_here_file_from_iso fatal},
    qw{get_temp_dir make_archive_with_files_called make_iuk_with_files},
    qw{run_as_root space_available stdout_as_root unpacked_size},
    qw{verify_signature}
);

use autodie qw(:all);
use Archive::Tar;
use Carp;
use Carp::Assert;
use Carp::Assert::More;
use Cwd;
use Data::Dumper;
use English qw{-no_match_vars};
use File::Temp qw{tempdir tempfile};
use GnuPG::Interface;
use Filesys::Df;
use IPC::System::Simple qw{capturex};
use List::Util qw{sum};
use Method::Signatures::Simple;
use Path::Class;


=head1 FUNCTIONS

=cut

func extract_file_from_iso($file, $iso) {
    my @cmd = qw{bsdtar -x --no-same-permissions --to-stdout --fast-read};
    push @cmd, ('--file', $iso, $file);
    open(my $cmd, '-|', @cmd);
    my $output = do { local $/; <$cmd> };
    close $cmd;
    "${^CHILD_ERROR_NATIVE}" == 0 or croak "bsdtar failed: ${^CHILD_ERROR_NATIVE}";
    return $output;
}

func extract_here_file_from_iso($dir, $iso) {
    my @cmd = qw{bsdtar -x --no-same-permissions};
    push @cmd, ('--file', $iso, $dir);
    system(@cmd);
    "${^CHILD_ERROR_NATIVE}" == 0 or croak "bsdtar failed: ${^CHILD_ERROR_NATIVE}";
    return;
}

func run_as_root(@command) {
    system("sudo", "-n", @command);
}

func stdout_as_root(@command) {
    capturex(qw{sudo -n}, @command);
}

sub get_temp_dir {
    dir(tempdir(CLEANUP => 0));
}

func fatal (%args) {
    assert(exists $args{msg});
    assert_isa($args{msg}, 'ARRAY');

    chdir '/';

    if (exists $args{rmtree} && defined $args{rmtree}) {
        if (exists $args{rmtree_as_root} && defined $args{rmtree_as_root} && $args{rmtree_as_root}) {
            run_as_root(
                qw{rm --recursive --force --preserve-root}, @{$args{rmtree}}
            );
        }
        else {
            foreach my $dir (@{$args{rmtree}}) {
                dir($dir)->rmtree;
            }
        }
    }

    croak(@{$args{msg}});
}

func unpacked_size ($archive_file) {
    assert(-e $archive_file);

    sum(0,
        map {
            my $ret = 0;
            if (my ($size) = (m{
                                   \A           # at the beginning of the string
                                   [-a-z]+      # permissions
                                   [[:space:]]+
                                   [[:digit:]]+ # owner
                                   /
                                   \d+          # group
                                   [[:space:]]+
                                   (\d+)        # size
                                   [[:space:]]+
                           }xms)) {
                $ret = $size;
            }
            $ret;
        } split(/\n/, `tar tv --numeric-owner -f $archive_file`)
    );
}

func make_archive_with_files_called ($archive_file, %args) {
    assert(exists $args{filenames});
    assert_isa($args{filenames}, 'ARRAY');
    assert_is(scalar(grep { exists $args{$_} } qw{size content}), 1);

    my $archive = Archive::Tar->new();
    my $orig_cwd = getcwd;
    chdir $archive_file->dir;
    foreach (@{$args{filenames}}) {
        file($_)->dir->mkpath;
        if (exists $args{content}) {
            my $fh = file($_)->openw;
            print $fh $args{content};
            close $fh;
        }
        elsif (exists $args{size}) {
            system("dd", "if=/dev/zero", "of=$_", "bs=1M", "count=".$args{size});
            ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to create ", $args{size}, "MB file '$_'.");
        }
        $archive->add_files($_);
    }
    $archive->write($archive_file)
        or croak("Failed to write '$archive_file':\n", $archive->error);
    chdir $orig_cwd;
}

func make_iuk_with_files ($iuk_filename, $tempdir, @files) {
    my $orig_cwd = getcwd;

    my $iuk = Archive::Tar->new();
    chdir $tempdir;

    my $fh = file('FORMAT')->openw; print $fh 1; close $fh;
    $iuk->add_files('FORMAT');
    unless (grep { $_ eq 'control.yml' } @files) {
        file('control.yml')->touch;
        $iuk->add_files('control.yml');
    }
    map { $iuk->add_files($_) or croak("Could not add file '$_'", $iuk->error) } @files;

    my $res = $iuk->write($iuk_filename);
    chdir $orig_cwd;
    $res or croak("Could not write IUK '$iuk_filename':\n", $iuk->error);
}

=head2 space_available

Returns the number of available bytes there are in directory $dir.

=cut
func space_available ($dir) {
    assert_defined($dir);

    my $df = df($dir, 1); # "1" means "please return the value in bytes"

    assert_defined($df);
    assert_exists($df, 'bavail');
    return $df->{bavail};
}

func verify_signature ($txt, $signature_txt, $trusted_gnupg_homedir) {
    assert_defined($txt);
    assert_nonblank($signature_txt);
    assert(-d $trusted_gnupg_homedir);

    my $gnupg = GnuPG::Interface->new();
    $gnupg->options->hash_init(
        homedir    => $trusted_gnupg_homedir,
        # We decide what key should be trusted by a given Tails,
        # and we won't put a key created in the future in there,
        # so if a key appears to be created in the future,
        # it must be because the clock has problems,
        # so we can ignore that.
        # Same for a key that appears to be expired.
        # Disable locking entirely: our GnuPG homedir is read-only.
        extra_args => [
            qw{--ignore-valid-from --ignore-time-conflict --lock-never}
        ],
    );

    my   ($signature_fh, $signature_file) = tempfile(CLEANUP => 1);
    print $signature_fh  $signature_txt;
    close $signature_fh;

    my   ($txt_fh,       $txt_file)       = tempfile(CLEANUP => 1);
    print $txt_fh        $txt;
    close $txt_fh;

    my ($stdout, $stderr) = (IO::Handle->new(), IO::Handle->new());
    my $pid = $gnupg->verify(
        handles => GnuPG::Handles->new(stdout => $stdout, stderr => $stderr),
        command_args => [ $signature_file, $txt_file ],
    );
    waitpid $pid, 0;

    return $CHILD_ERROR == 0;
}

1;
